import json
import itertools

import SCons
from SCons.Builder import Builder, DictEmitter, ListEmitter
from SCons.Action import Action
from SCons.Script import GetOption, AddOption


def enable(env, config):
    compile_commands = {}
    merged_compile_commands = {}
    db_entry_nodes = []

    env['_COMPILE_DB_ID'] = id(compile_commands)

    def create_db_entry_emitter(cxx, shared):
        def emitter(target, source, env):
            if env.get('_COMPILE_DB_ID') != id(compile_commands):
                return target, source

            def add_db_entry():
                entry = config.entry_func(env, target, source, cxx, shared)
                if entry:
                    key = '{}:{}'.format(entry['file'], str(target[0]))
                    compile_commands[key] = entry

            entry_node = SCons.Node.Python.Value(source)
            entry = env._AddDbEntry(entry_node, [],
                                    _COMPILE_DB_ENTRY_FUNC=add_db_entry)
            env.AlwaysBuild(entry)
            env.NoCache(entry)
            db_entry_nodes.append(entry_node)
            return target, source
        return emitter

    def add_db_entry_action(target, source, env):
        env['_COMPILE_DB_ENTRY_FUNC']()

    def update_db_action(target, source, env):
        # Convert dict to a list sorted with file/output tuple.
        contents = [e for _, e in sorted(merged_compile_commands.items())]
        with open(target[0].path, 'w') as f:
            json.dump(contents, f, indent=2)

    def update_internal_db_action(target, source, env):
        merged_compile_commands.update(compile_commands)
        with open(target[0].path, 'w') as f:
            json.dump(merged_compile_commands, f, sort_keys=True)

    #
    # Hook new emitters to the existing ones
    #
    for ((cxx, suffix), shared) in itertools.product(
            [(True, s) for s in config.cxx_suffixes] +
            [(False, s) for s in config.cc_suffixes],
            (True, False)):
        builder = 'SharedObject' if shared else 'StaticObject'
        emitter = env['BUILDERS'][builder].emitter
        assert isinstance(emitter, DictEmitter)
        org = emitter[suffix]
        new = create_db_entry_emitter(cxx, shared)
        emitter[suffix] = ListEmitter((org, new))

    #
    # Add builders
    #
    env['BUILDERS']['_AddDbEntry'] = Builder(
        action=Action(add_db_entry_action, None))

    env['BUILDERS']['_UpdateInternalDb'] = Builder(
        action=Action(update_internal_db_action,
                      'Check compilation DB : $TARGET'),
        target_scanner=SCons.Scanner.Scanner(
            function=lambda node, env, path: db_entry_nodes,
            node_class=None))

    env['BUILDERS']['_UpdateDb'] = Builder(
        action=Action(update_db_action, 'Update compilation DB: $TARGET'))

    def compile_db(env, target=config.db):
        merged_compile_commands.clear()
        merged_db = env._UpdateInternalDb('{}.internal'.format(target), [])[0]
        if merged_db.exists():
            merged_compile_commands.update(
                json.loads(merged_db.get_text_contents()))
        env.AlwaysBuild(merged_db)
        return env._UpdateDb(target, merged_db)
    env.AddMethod(compile_db, 'CompileDb')


def enable_with_cmdline(env, config, option_name, option_help):
    def is_option_on():
        def add_script_option():
            AddOption('--' + option_name, dest='compile_db',
                      action='store_true', help=option_help)
        try:
            return GetOption('compile_db')
        except AttributeError:
            add_script_option()
            return GetOption('compile_db')

    if not is_option_on():
        return

    enable(env, config)
    db = env.CompileDb(config.db)
    env.Default(env.Alias('compiledb', db))


def enabled(env):
    return '_UpdateDb' in env['BUILDERS']



from . import core, entry_func

__version__ = '0.5.1'

DEFAULT_OPTION_NAME = 'compiledb'
DEFAULT_DB_NAME = 'compile_commands.json'


def enable(env, config=None):
    """
    Enable compilation command collection by hooking into object builders.

    enable() adds a new builder, CompileDb to generates DB file. CompileDb
    builder can specify the target DB file and the default is
    'compile_commands.json'.

    Note that CompileDb builder only collects info for the object builders
    since enable() is called.
    """
    return core.enable(env, config if config else Config())


def enable_with_cmdline(env, config=None,
                        option_name=DEFAULT_OPTION_NAME,
                        option_help='Update {}'.format(DEFAULT_DB_NAME)):
    """
    Enable and build DB with command line option, --compiledb.
    """
    return core.enable_with_cmdline(
        env, config if config else Config(), option_name, option_help)


entry_func_simple = entry_func.simple
entry_func_default = entry_func.default


class Config:
    """
    Customise the overall behaviour.

    - db           : file name of compilation DB.
    - cxx_suffixes : Suffixes for C++ files. Default is ('.cpp', and '.cc').
    - cc_suffixes  : Suffixes if C files. Default is ('.c',).
    - entry_func   : a function to determine the entry dict for each file.
        Predefined functions or a user-defined function can be used.
        It includes:

        - 'entry_func_default': This is default and it should work for most of
          cases

        - 'entry_func_simple': Use CPPPATH and CPPDEFINES only and
          'clang'/'clang++' as compiler tool name. This will be useful to use
          clangd with compilers, which use command line arguments clangd cannot
          understand.
    """

    def __init__(self,
                 db=DEFAULT_DB_NAME,
                 cxx_suffixes=('.cpp', '.cc'),
                 cc_suffixes=('.c',),
                 entry_func=entry_func_default):
        self.db = db
        self.cc_suffixes = cc_suffixes
        self.cxx_suffixes = cxx_suffixes
        self.entry_func = entry_func


def enabled(env):
    return core.enabled(env)



from SCons.Action import Action


def default(env, target, source, cxx, shared):
    """
    A default DB entry function which will work for the most of cases.
    """
    cmdstr = '${}{}'.format('SH' if shared else '',
                            'CXXCOM' if cxx else 'CCCOM')
    command = Action(cmdstr).strfunction(target, source, env)
    return {'directory': env.Dir('#').abspath,
            'file': str(source[0]),
            'command': command}


def simple(env, target, source, cxx, shared):
    """
    A simple DB entry function to pretend that the current tool chain uses
    clang/clang++.
    """
    flags = '{} {}'.format(
        ' '.join('-D{}'.format(d) for d in env.get('CPPDEFINES')),
        ' '.join('-I{}'.format(str(env.Dir(p))) for p in env.get('CPPPATH')))

    source_path = str(source[0])
    toolchain = 'clang++' if cxx else 'clang'
    return {'directory': env.Dir('#').abspath,
            'file': source_path,
            'command': '{} {} -c {}'.format(toolchain, flags, source_path)}



import setuptools
import re

with open('README.md', 'rt') as f:
    long_description = f.read()

with open('scons_compiledb/__init__.py', "rt") as f:
    version = re.search(r'__version__ = \'(.*?)\'', f.read()).group(1)

setuptools.setup(
    name='scons-compiledb',
    version=version,
    author='Hans Jang',
    author_email='hsjang8848@gmail.com',
    description='compile_commands.json generation for SCons build system',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/pinetr2e/scons-compiledb',
    packages=setuptools.find_packages(),
    setup_requires=['pytest-runner'],
    tests_require=["pytest"],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        'Intended Audience :: Developers',
        'Environment :: Console',
        'Topic :: Software Development',
    ],
)
